
import logging
import pickle
import random
import sys
import threading
import time

import pytest

import locked_dict.locked_dict as locked_dict


def test_locked_dict():
    """Simple test driver. Switch to debug logging, if any argument given."""

    _start = time.time()

    a_level = logging.INFO if len(sys.argv) < 2 else logging.DEBUG
    a_format = '%(asctime)s: %(message)s'

    logging.basicConfig(format=a_format)
    logger = logging.getLogger(__name__)
    logger.setLevel(a_level)

    def worker(key_seq, shared_map):
        for key in key_seq:
            shared_map[key] = {threading.currentThread().getName(): key}
            logger.debug('{} {}'.format(key, shared_map[key]))
            pause = float(random.randint(1, 5)) / 1000.
            logger.debug('sleeping %02.3f' % (pause,))
            time.sleep(pause)

        return len(key_seq)

    expected = 0
    d = locked_dict.LockedDict()
    with d as m:
        m[0] = ['foo']
        expected += 1
        m.clear()
        expected -= 1
        assert len(m) == expected
        try:
            # noinspection PyUnusedLocal
            __ = m.popitem()
            assert False
        except KeyError:
            pass
        try:
            del m['not_there']
            assert False
        except KeyError:
            pass
        m.update({0: 'foo'})
        expected += 1
        assert len(m) == expected

    logger.debug('Entries({:5d}/{:5d})'.format(len(d), expected))

    for k, v in d.items():
        logger.debug("{}: {}".format(k, v))
    logger.debug('{} {} {}'.format(
        id(d), isinstance(d, dict), isinstance(d, locked_dict.LockedDict)))
    logger.debug(dir(d))
    logger.debug(dir(getattr(d, '_lock')))

    with d as m:
        d_ser = pickle.dumps(m)
    logger.debug(d_ser)

    with d as m:
        m[1] = ('bar',)
        expected += 1
        m[42] = {'baz': 'ooka'}
        expected += 1
        x = m.pop(42)
        expected -= 1
        m[42] = x
        expected += 1
        __ = m.get(-42)
        assert __ is None
        if -42 not in m:
            __ = locked_dict.LockedDict.fromkeys(m.keys(), 'yes')
            bf = dict([(z, 'yes') for z in m.keys()])
            assert __ == bf
            assert __ is not bf
            logger.debug(__)

    with d as m:
        d_ser = pickle.dumps(m)
    logger.debug(d_ser)

    rd = pickle.loads(d_ser)
    for k, v in rd.items():
        logger.debug("{}: {}".format(k, v))

    logger.debug('Entries({:5d}/{:5d})'.format(len(rd), expected))

    logger.debug('{} {} {}'.format(
        id(rd), type(rd), isinstance(rd, (dict, locked_dict.LockedDict))))

    worker_tasks = random.randint(1, 234)
    worker_count = random.randint(5, 67)
    logger.debug("Starting {} workers on {} tasks each ..."
                 "".format(worker_count, worker_tasks))
    expected += worker_count * worker_tasks
    for i in range(43, 43 + worker_count * worker_tasks, worker_tasks):
        t = threading.Thread(
            target=worker, args=(range(i, i + worker_tasks), rd))
        t.setDaemon(True)
        t.start()

    main_thread = threading.currentThread()
    for t in threading.enumerate():
        if t is main_thread:
            continue
        logger.debug('joining %s' % (t.getName(),))
        t.join()

    for k, v in rd.items():
        logger.debug("{}: {}".format(k, v))
    assert len(rd) == expected
    logger.info(
        'WorkersTasks({:2d}:{:3d}).Entries({:5d}/{:5d}); SizeBytes({:6d})'
        ' - with Python({}); LatencySecs({:0.3f})'
        ''.format(worker_count, worker_tasks, len(rd), expected,
                  sys.getsizeof(rd),
                  ','.join(["{:2d}".format(z) for z in sys.version_info[:3]]),
                  round(time.time() - _start, 3)))


def test_main():
    assert locked_dict  # use your library here


def test_lock_empty():
    expected = 0
    d = locked_dict.LockedDict()
    assert len(d) == expected
    assert bool(d) is False
    assert d is not True
    assert hasattr(d, '_lock')
    empty_d = {}
    assert d == empty_d
    d.setdefault('not_there_yet', []).append(42)
    assert d['not_there_yet'] == [42]


def test_lock_with():
    expected = 0
    d = locked_dict.LockedDict()
    empty_d = {}
    plain_old_d = {999: 'plain old dict', 12345: 54321}
    assert d != plain_old_d

    with d as m:
        assert len(m) == expected
        assert bool(m) is False
        assert m is not True
        assert hasattr(m, '_lock')
        assert m != plain_old_d
        assert m == empty_d

        m[0] = ['foo']
        expected += 1
        assert len(m) == expected
        assert bool(m) is True
        assert m is not False
        assert m != plain_old_d
        assert m != empty_d

        m.clear()
        expected -= 1
        assert len(m) == expected
        assert bool(m) is False
        assert m is not True
        assert m != plain_old_d
        assert m == empty_d

        with pytest.raises(KeyError):
            # noinspection PyUnusedLocal                                                                                                                                 
            __ = m.popitem()

        with pytest.raises(KeyError):
            del m['not_there']

        m.update({0: 'foo'})
        expected += 1
        assert len(m) == expected
        assert bool(m) is True
        assert m is not False
        assert m != plain_old_d
        assert m != empty_d

        d_foo_ser = pickle.dumps(m[0])
        del m[0]
        assert m.get(0) is None
        m[0] = pickle.loads(d_foo_ser)
        assert m[0] == 'foo'



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""This module provides a single class derived from dict to
allow thread safe and mutable iterations through a lock. """

import itertools
import threading

__all__ = ['LockedDict']


# Use basestring name lookup test to adapt to python version
try:
    # noinspection PyCompatibility
    base_str = basestring
    items = 'iteritems'  # from here on we are python < v3
except NameError:  # ... below section applies to python v3+
    base_str = str, bytes, bytearray
    items = 'items'


# noinspection PyShadowingBuiltins
class LockedDict(dict):
    """Work horse like dict derivative with an added lock to allow for
    context managed thread safe operation sections.
    Some methods added, so instances mostly behave like real dicts.
    Instances can be (de)serialized with pickle methods, given the
    lock state - skipped in dump and reload operations - is stable.
    Due to items attribute abstraction works with python v2.7+.
    """

    __slots__ = ('_lock',)  # no __dict__ - that would be redundant

    @staticmethod
    def _process_args(map_or_it=(), **kwargs):
        """Custom made helper for this class."""
        if hasattr(map_or_it, items):
            map_or_it = getattr(map_or_it, items)()
        it_chain = itertools.chain
        return ((k, v)
                for k, v in it_chain(map_or_it, getattr(kwargs, items)()))

    def __init__(self, mapping=(), **kwargs):
        """Base (dict) accepts mappings or iterables as first argument."""
        super(LockedDict, self).__init__(self._process_args(mapping, **kwargs))
        self._lock = threading.Lock()

    def __enter__(self):
        """Context manager enter the block, acquire the lock."""
        self._lock.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit the block, release the lock."""
        self._lock.release()

    def __getstate__(self):
        """Enable Pickling inside context blocks,
        through inclusion of the slot entries without the lock."""
        return dict(
            (slot, getattr(self, slot))
            for slot in self.__slots__
            if hasattr(self, slot) and slot != '_lock'
        )

    def __setstate__(self, state):
        """Restore the instance from pickle including the slot entries,
        without addition of a fresh lock.
        """
        for slot, value in getattr(state, items)():
            setattr(self, slot, value)
        self._lock = threading.Lock()

    def __getitem__(self, k):
        """For now plain delegation of getitem method to base class."""
        return super(LockedDict, self).__getitem__(k)

    def __setitem__(self, k, v):
        """For now plain delegation of setitem method to base class."""
        return super(LockedDict, self).__setitem__(k, v)

    def __delitem__(self, k):
        """For now plain delegation of del method to base class."""
        return super(LockedDict, self).__delitem__(k)

    def get(self, k, default=None):
        """For now plain delegation of get method to base class."""
        return super(LockedDict, self).get(k, default)

    def setdefault(self, k, default=None):
        """For now plain delegation of setdefault method to base class."""
        return super(LockedDict, self).setdefault(k, default)

    def pop(self, k, d=None):
        """For now plain delegation of pop method to base class."""
        return super(LockedDict, self).pop(k, d)

    def update(self, map_or_it=(), **kwargs):
        """Ensure processing of mappings or iterables as first argument."""
        super(LockedDict, self).update(self._process_args(map_or_it, **kwargs))

    def __contains__(self, k):
        """For now plain delegation of contains method to base class."""
        return super(LockedDict, self).__contains__(k)

    # noinspection PyMethodOverriding
    @classmethod
    def fromkeys(cls, seq, value=None):
        """For now plain delegation of fromkeys class method to base."""
        return super(LockedDict, cls).fromkeys(seq, value)
